"""
Multi-agent predator-prey environment

Multi-agent environment where a Flock of predator
type agents attempt to capture a flock of prey
agents.
"""
from . import env, generator, observations, rewards, types, viewer
from .env import PredatorPrey
