"""
Flock mult-agent environment

Multi-agent environment inspired by Reynolds boid model. Agents
aim to stay close to other agents without colliding.
"""
from floxs.flock import env, generator, observations, rewards, types, viewer
from floxs.flock.env import Flock
